@@ -1,4 +1,4 @@
-# Copyright 2018-2020 Jérôme Dumonteil
+# Copyright 2018-2024 Jérôme Dumonteil
 # Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -19,9 +19,10 @@
 # https://github.com/lpod/lpod-python
 # Authors: David Versmisse <david.versmisse@itaapy.com>
 #          Hervé Cauwelier <herve@itaapy.com>
-"""Constants, XML tags and ODF headers
+"""Constants, XML tags and ODF headers.
 """
-ODF_TEMPLATES_DIR = "templates"
+from __future__ import annotations
+
 ODF_TEMPLATES = {
     "text": "text.ott",
     "texte": "text.ott",
@@ -107,6 +108,59 @@ ODF_META = "meta.xml"
 ODF_SETTINGS = "settings.xml"
 ODF_STYLES = "styles.xml"
 ODF_MANIFEST = "META-INF/manifest.xml"
+ODF_MANIFEST_NAME = "manifest.xml"
+
+OFFICE_VERSION = "1.1"
+
+OFFICE_PREFIX = b"""<?xml version="1.0" encoding="UTF-8"?>
+<office:document
+  xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0"
+  xmlns:style="urn:oasis:names:tc:opendocument:xmlns:style:1.0"
+  xmlns:text="urn:oasis:names:tc:opendocument:xmlns:text:1.0"
+  xmlns:table="urn:oasis:names:tc:opendocument:xmlns:table:1.0"
+  xmlns:draw="urn:oasis:names:tc:opendocument:xmlns:drawing:1.0"
+  xmlns:fo="urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0"
+  xmlns:xlink="http://www.w3.org/1999/xlink"
+  xmlns:dc="http://purl.org/dc/elements/1.1/"
+  xmlns:meta="urn:oasis:names:tc:opendocument:xmlns:meta:1.0"
+  xmlns:number="urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0"
+  xmlns:presentation="urn:oasis:names:tc:opendocument:xmlns:presentation:1.0"
+  xmlns:svg="urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0"
+  xmlns:chart="urn:oasis:names:tc:opendocument:xmlns:chart:1.0"
+  xmlns:dr3d="urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0"
+  xmlns:math="http://www.w3.org/1998/Math/MathML"
+  xmlns:form="urn:oasis:names:tc:opendocument:xmlns:form:1.0"
+  xmlns:script="urn:oasis:names:tc:opendocument:xmlns:script:1.0"
+  xmlns:ooo="http://openoffice.org/2004/office"
+  xmlns:oooc="http://openoffice.org/2004/calc"
+  xmlns:ooow="http://openoffice.org/2004/writer"
+  xmlns:xforms="http://www.w3.org/2002/xforms"
+  xmlns:xsd="http://www.w3.org/2001/XMLSchema"
+  xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
+  xmlns:smil="urn:oasis:names:tc:opendocument:xmlns:smil-compatible:1.0"
+  xmlns:anim="urn:oasis:names:tc:opendocument:xmlns:animation:1.0"
+  xmlns:rpt="http://openoffice.org/2005/report"
+  xmlns:of="urn:oasis:names:tc:opendocument:xmlns:of:1.2"
+  xmlns:xhtml="http://www.w3.org/1999/xhtml"
+  xmlns:grddl="http://www.w3.org/2003/g/data-view#"
+  xmlns:officeooo="http://openoffice.org/2009/office"
+  xmlns:tableooo="http://openoffice.org/2009/table"
+  xmlns:drawooo="http://openoffice.org/2010/draw"
+  xmlns:calcext="urn:org:documentfoundation:names:experimental:calc:xmlns:calcext:1.0"
+  xmlns:loext="urn:org:documentfoundation:names:experimental:office:xmlns:loext:1.0"
+  xmlns:field="urn:openoffice:names:experimental:ooo-ms-interop:xmlns:field:1.0"
+  xmlns:formx="urn:openoffice:names:experimental:ooxml-odf-interop:xmlns:form:1.0"
+  xmlns:css3t="http://www.w3.org/TR/css3-text/"
+  xmlns:config="urn:oasis:names:tc:opendocument:xmlns:config:1.0"
+  xmlns:dom="http://www.w3.org/2001/xml-events"
+  xmlns:manifest="urn:oasis:names:tc:opendocument:xmlns:manifest:1.0"
+  xmlns:rdfa="http://docs.oasis-open.org/opendocument/meta/rdfa#"
+"""
+# packaging
+FOLDER = "folder"
+ZIP = "zip"
+XML = "xml"
+PACKAGING = {FOLDER, XML, ZIP}
 
 # Presentation classes (for layout)
 ODF_CLASSES = (
@@ -555,6 +609,25 @@ ODF_PROPERTIES = {
     "text:space-before",
 }
 
+ODF_COLOR_PROPERTY = {
+    "dr3d:ambient-color",
+    "dr3d:diffuse-color",
+    "dr3d:emissive-color",
+    "dr3d:specular-color",
+    "draw:fill-color",
+    "draw:secondary-fill-color",
+    "draw:shadow-color",
+    "fo:background-color",
+    "fo:color",
+    "smil:fadeColor",
+    "style:layout-grid-color",
+    "style:text-line-through-color",
+    "style:text-overline-color",
+    "style:text-underline-color",
+    "style:use-window-font-color",
+    "svg:stroke-color",
+}
+
 # from CSS3 color map
 CSS3_COLORMAP = {
     "indigo": (75, 0, 130),
