@@ -1,4 +1,4 @@
-# Copyright 2018-2020 Jérôme Dumonteil
+# Copyright 2018-2024 Jérôme Dumonteil
 # Copyright (c) 2009-2013 Ars Aperta, Itaapy, Pierlis, Talend.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -20,31 +20,45 @@
 # Authors: David Versmisse <david.versmisse@itaapy.com>
 #          Hervé Cauwelier <herve@itaapy.com>
 #          Jerome Dumonteil <jerome.dumonteil@itaapy.com>
-"""Meta class for meta.xml part
+"""Meta class for meta.xml part.
 """
-from datetime import timedelta, datetime
+from __future__ import annotations
+
 from datetime import date as dtdate
+from datetime import datetime, timedelta
 from decimal import Decimal
 from string import ascii_letters, digits
+from typing import Any
 
-from .datatype import DateTime, Duration, Boolean, Date
+from .datatype import Boolean, Date, DateTime, Duration
 from .element import Element
-from .xmlpart import XmlPart
+from .meta_auto_reload import MetaAutoReload
+from .meta_hyperlink_behaviour import MetaHyperlinkBehaviour
+from .meta_template import MetaTemplate
+from .mixin_dc_creator import DcCreatorMixin
+from .mixin_dc_date import DcDateMixin
 from .utils import to_str
 from .version import __version__
+from .xmlpart import XmlPart
+
+GENERATOR = f"odfdo {__version__}"
 
 
-class Meta(XmlPart):
-    _generator_modified = False
+class Meta(XmlPart, DcCreatorMixin, DcDateMixin):
+    def __init__(self, *args: Any, **kwargs: Any) -> None:
+        super().__init__(*args, **kwargs)
+        self._generator_modified: bool = False
 
-    def get_meta_body(self):
+    def get_meta_body(self) -> Element:
         return self.get_element("//office:meta")
 
-    def get_title(self):
+    def get_title(self) -> str | None:
         """Get the title of the document.
 
         This is not the first heading but the title metadata.
 
+        (Also available as "self.title" property.)
+
         Return: str (or None if inexistant)
         """
         element = self.get_element("//dc:title")
@@ -52,11 +66,13 @@ class Meta(XmlPart):
             return None
         return element.text
 
-    def set_title(self, title):
+    def set_title(self, title: str) -> None:
         """Set the title of the document.
 
         This is not the first heading but the title metadata.
 
+        (Also available as "self.title" property.)
+
         Arguments:
 
             title -- str
@@ -67,9 +83,23 @@ class Meta(XmlPart):
             self.get_meta_body().append(element)
         element.text = title
 
-    def get_description(self):
+    @property
+    def title(self) -> str | None:
+        """Get or set the title of the document <dc:title>.
+
+        Return: str (or None if inexistant)
+        """
+        return self.get_title()
+
+    @title.setter
+    def title(self, title: str) -> None:
+        return self.set_title(title)
+
+    def get_description(self) -> str | None:
         """Get the description of the document. Also known as comments.
 
+        (Also available as "self.description" property.)
+
         Return: str (or None if inexistant)
         """
         element = self.get_element("//dc:description")
@@ -80,9 +110,11 @@ class Meta(XmlPart):
     # As named in OOo
     get_comments = get_description
 
-    def set_description(self, description):
+    def set_description(self, description: str) -> None:
         """Set the description of the document. Also known as comments.
 
+        (Also available as "self.description" property.)
+
         Arguments:
 
             description -- str
@@ -95,9 +127,23 @@ class Meta(XmlPart):
 
     set_comments = set_description
 
-    def get_subject(self):
+    @property
+    def description(self) -> str | None:
+        """Get or set the description of a document <dc:description>.
+
+        Return: str (or None if inexistant)
+        """
+        return self.get_description()
+
+    @description.setter
+    def description(self, description: str) -> None:
+        return self.set_description(description)
+
+    def get_subject(self) -> str | None:
         """Get the subject of the document.
 
+        (Also available as "self.subject" property.)
+
         Return: str (or None if inexistant)
         """
         element = self.get_element("//dc:subject")
@@ -105,9 +151,11 @@ class Meta(XmlPart):
             return None
         return element.text
 
-    def set_subject(self, subject):
+    def set_subject(self, subject: str) -> None:
         """Set the subject of the document.
 
+        (Also available as "self.subject" property.)
+
         Arguments:
 
             subject -- str
@@ -118,14 +166,28 @@ class Meta(XmlPart):
             self.get_meta_body().append(element)
         element.text = subject
 
-    def get_language(self):
-        """Get the language code of the document.
+    @property
+    def subject(self) -> str | None:
+        """Get or set the subject of a document <dc:subject>.
+
+        Return: str (or None if inexistant)
+        """
+        return self.get_subject()
+
+    @subject.setter
+    def subject(self, subject: str) -> None:
+        return self.set_subject(subject)
+
+    def get_language(self) -> str | None:
+        """Get the default language of the document.
+
+        (Also available as "self.language" property.)
 
         Return: str (or None if inexistant)
 
         Example::
 
-            >>> document.get_language()
+            >>> document.meta.get_language()
             fr-FR
         """
         element = self.get_element("//dc:language")
@@ -133,8 +195,10 @@ class Meta(XmlPart):
             return None
         return element.text
 
-    def set_language(self, language):
-        """Set the language code of the document.
+    def set_language(self, language: str) -> None:
+        """Set the default language of the document.
+
+        (Also available as "self.language" property.)
 
         Arguments:
 
@@ -142,11 +206,13 @@ class Meta(XmlPart):
 
         Example::
 
-            >>> document.set_language('fr-FR')
+            >>> document.meta.set_language('fr-FR')
         """
-        language = to_str(language)
+        language = str(language)
         if not self._is_RFC3066(language):
-            raise TypeError('language must be "xx-YY" lang-COUNTRY code (RFC3066)')
+            raise TypeError(
+                'Language must be "xx" lang or "xx-YY" lang-COUNTRY code (RFC3066)'
+            )
         element = self.get_element("//dc:language")
         if element is None:
             element = Element.from_tag("dc:language")
@@ -154,20 +220,14 @@ class Meta(XmlPart):
         element.text = language
 
     @staticmethod
-    def _is_RFC3066(lang):
-        def test_part1(p1):
-            if not 2 <= len(p1) <= 3:
+    def _is_RFC3066(lang: str) -> bool:
+        def test_part1(part1: str) -> bool:
+            if not 2 <= len(part1) <= 3:
                 return False
-            for x in p1:
-                if x not in ascii_letters:
-                    return False
-            return True
+            return all(x in ascii_letters for x in part1)
 
-        def test_part2(p1):
-            for x in p1:
-                if x not in ascii_letters and x not in digits:
-                    return False
-            return True
+        def test_part2(part2: str) -> bool:
+            return all(x in ascii_letters or x in digits for x in part2)
 
         if not lang or not isinstance(lang, str):
             return False
@@ -178,38 +238,25 @@ class Meta(XmlPart):
             return False
         if not test_part1(parts[0]):
             return False
-        for p in parts[1:]:
-            if not test_part2(p):
-                return False
-        return True
+        return all(test_part2(p) for p in parts[1:])
 
-    def get_modification_date(self):
-        """Get the last modified date of the document.
+    @property
+    def language(self) -> str | None:
+        """Get or set the default language of the document <dc:language>.
 
-        Return: datetime (or None if inexistant)
+        Return: str (or None if inexistant)
         """
-        element = self.get_element("//dc:date")
-        if element is None:
-            return None
-        modification_date = element.text
-        return DateTime.decode(modification_date)
+        return self.get_language()
 
-    def set_modification_date(self, date):
-        """Set the last modified date of the document.
+    @language.setter
+    def language(self, language: str) -> None:
+        return self.set_language(language)
 
-        Arguments:
-
-            date -- datetime
-        """
-        element = self.get_element("//dc:date")
-        if element is None:
-            element = Element.from_tag("dc:date")
-            self.get_meta_body().append(element)
-        element.text = DateTime.encode(date)
-
-    def get_creation_date(self):
+    def get_creation_date(self) -> datetime | None:
         """Get the creation date of the document.
 
+        (Also available as "self.creation_date" property.)
+
         Return: datetime (or None if inexistant)
         """
         element = self.get_element("//meta:creation-date")
@@ -218,9 +265,13 @@ class Meta(XmlPart):
         creation_date = element.text
         return DateTime.decode(creation_date)
 
-    def set_creation_date(self, date):
+    def set_creation_date(self, date: datetime | None = None) -> None:
         """Set the creation date of the document.
 
+        If provided datetime is None, use current time.
+
+        (Also available as "self.creation_date" property.)
+
         Arguments:
 
             date -- datetime
@@ -229,16 +280,131 @@ class Meta(XmlPart):
         if element is None:
             element = Element.from_tag("meta:creation-date")
             self.get_meta_body().append(element)
+        if date is None:
+            date = datetime.now()
         element.text = DateTime.encode(date)
 
-    def get_initial_creator(self):
+    @property
+    def creation_date(self) -> datetime | None:
+        """Get or set the date and time when a document was created
+        <meta:creation-date>.
+
+        If provided datetime is None, use current time.
+
+        Return: datetime (or None if inexistant)
+        """
+        return self.get_creation_date()
+
+    @creation_date.setter
+    def creation_date(self, date: datetime | None = None) -> None:
+        return self.set_creation_date(date)
+
+    @property
+    def print_date(self) -> datetime | None:
+        """Get or set the date and time when a document when a document was last printed
+        <meta:print-date>
+
+        If provided datetime is None, use current time.
+
+        Return: datetime (or None if inexistant)
+        """
+        element = self.get_element("//meta:print-date")
+        if element is None:
+            return None
+        date = element.text
+        return DateTime.decode(date)
+
+    @print_date.setter
+    def print_date(self, date: datetime | None = None) -> None:
+        element = self.get_element("//meta:print-date")
+        if element is None:
+            element = Element.from_tag("meta:print-date")
+            self.get_meta_body().append(element)
+        if date is None:
+            date = datetime.now()
+        element.text = DateTime.encode(date)
+
+    def get_template(self) -> MetaTemplate | None:
+        """Get the MetaTemplate <meta:template> element or None."""
+        element = self.get_element("//meta:template")
+        if element is None:
+            return None
+        return element
+
+    @property
+    def template(self) -> MetaTemplate | None:
+        """Get the MetaTemplate <meta:template> element or None."""
+        return self.get_template()
+
+    def set_template(
+        self,
+        date: datetime | None = None,
+        href: str = "",
+        title: str = "",
+    ) -> None:
+        """Set the MetaTemplate <meta:template> element."""
+        template = MetaTemplate(date=date, href=href, title=title)
+        current = self.template
+        if isinstance(current, MetaTemplate):
+            current.delete()
+        self.get_meta_body().append(template)
+
+    def get_auto_reload(self) -> MetaAutoReload | None:
+        """Get the MetaAutoReload <meta:auto-reload> element or None."""
+        element = self.get_element("//meta:auto-reload")
+        if element is None:
+            return None
+        return element
+
+    @property
+    def auto_reload(self) -> MetaAutoReload | None:
+        """Get the MetaAutoReload <meta:auto-reload> element or None."""
+        return self.get_auto_reload()
+
+    def set_auto_reload(self, delay: timedelta, href: str = "") -> None:
+        """Set the MetaAutoReload <meta:auto-reload> element."""
+        autoreload = MetaAutoReload(delay=delay, href=href)
+        current = self.auto_reload
+        if isinstance(current, MetaAutoReload):
+            current.delete()
+        self.get_meta_body().append(autoreload)
+
+    def get_hyperlink_behaviour(self) -> MetaHyperlinkBehaviour | None:
+        """Get the MetaHyperlinkBehaviour <meta:hyperlink-behaviour> element or None."""
+        element = self.get_element("//meta:hyperlink-behaviour")
+        if element is None:
+            return None
+        return element
+
+    @property
+    def hyperlink_behaviour(self) -> MetaAutoReload | None:
+        """Get the MetaHyperlinkBehaviour <meta:hyperlink-behaviour> element or None."""
+        return self.get_hyperlink_behaviour()
+
+    def set_hyperlink_behaviour(
+        self,
+        target_frame_name: str = "_blank",
+        show: str = "replace",
+    ) -> None:
+        """Set the MetaHyperlinkBehaviour <meta:hyperlink-behaviour> element."""
+        behaviour = MetaHyperlinkBehaviour(
+            target_frame_name=target_frame_name, show=show
+        )
+        current = self.hyperlink_behaviour
+        if isinstance(current, MetaHyperlinkBehaviour):
+            current.delete()
+        self.get_meta_body().append(behaviour)
+
+    def get_initial_creator(self) -> str | None:
         """Get the first creator of the document.
 
+        (Also available as "self.initial_creator" property.)
+
         Return: str (or None if inexistant)
 
         Example::
 
-            >>> document.get_initial_creator()
+            >>> document.meta.get_initial_creator()
             Unknown
         """
         element = self.get_element("//meta:initial-creator")
@@ -246,16 +412,18 @@ class Meta(XmlPart):
             return None
         return element.text
 
-    def set_initial_creator(self, creator):
+    def set_initial_creator(self, creator: str) -> None:
         """Set the first creator of the document.
 
+        (Also available as "self.initial_creator" property.)
+
         Arguments:
 
             creator -- str
 
         Example::
 
-            >>> document.set_initial_creator(u"Plato")
+            >>> document.meta.set_initial_creator("Plato")
         """
         element = self.get_element("//meta:initial-creator")
         if element is None:
@@ -263,42 +431,45 @@ class Meta(XmlPart):
             self.get_meta_body().append(element)
         element.text = creator
 
-    def get_creator(self):
-        """Get the creator of the document.
+    @property
+    def initial_creator(self) -> str | None:
+        """Get or set the initial creator of a document
+        <meta:initial-creator>.
 
         Return: str (or None if inexistant)
-
-        Example::
-
-            >>> document.get_creator()
-            Unknown
         """
-        element = self.get_element("//dc:creator")
+        return self.get_initial_creator()
+
+    @initial_creator.setter
+    def initial_creator(self, creator: str) -> None:
+        return self.set_initial_creator(creator)
+
+    @property
+    def printed_by(self) -> str | None:
+        """Get or set the name of the last person who printed a document.
+        <meta:printed-by>
+
+        Return: str (or None if inexistant)
+        """
+        element = self.get_element("//meta:printed-by")
         if element is None:
             return None
         return element.text
 
-    def set_creator(self, creator):
-        """Set the creator of the document.
-
-        Arguments:
-
-            creator -- str
-
-        Example::
-
-            >>> document.set_creator(u"Plato")
-        """
-        element = self.get_element("//dc:creator")
+    @printed_by.setter
+    def printed_by(self, printed_by: str) -> None:
+        element = self.get_element("//meta:printed-by")
         if element is None:
-            element = Element.from_tag("dc:creator")
+            element = Element.from_tag("meta:printed-by")
             self.get_meta_body().append(element)
-        element.text = creator
+        element.text = printed_by
 
-    def get_keywords(self):
+    def get_keywords(self) -> str | None:
         """Get the keywords of the document. Return the field as-is, without
         any assumption on the keyword separator.
 
+        (Also available as "self.keyword" and "self.keywords" property.)
+
         Return: str (or None if inexistant)
         """
         element = self.get_element("//meta:keyword")
@@ -306,10 +477,12 @@ class Meta(XmlPart):
             return None
         return element.text
 
-    def set_keywords(self, keywords):
+    def set_keywords(self, keywords: str) -> None:
         """Set the keywords of the document. Although the name is plural, a
         str string is required, so join your list first.
 
+        (Also available as "self.keyword" and "self.keywords" property.)
+
         Arguments:
 
             keywords -- str
@@ -320,10 +493,27 @@ class Meta(XmlPart):
             self.get_meta_body().append(element)
         element.text = keywords
 
-    def get_editing_duration(self):
+    @property
+    def keyword(self) -> str | None:
+        """Get or set some keyword(s) keyword pertaining to a document
+        <dc:keyword>.
+
+        Return: str (or None if inexistant)
+        """
+        return self.get_keywords()
+
+    @keyword.setter
+    def keyword(self, keyword: str) -> None:
+        return self.set_keywords(keyword)
+
+    keywords = keyword
+
+    def get_editing_duration(self) -> timedelta | None:
         """Get the time the document was edited, as reported by the
         generator.
 
+        (Also available as "self.editing_duration" property.)
+
         Return: timedelta (or None if inexistant)
         """
         element = self.get_element("//meta:editing-duration")
@@ -332,9 +522,11 @@ class Meta(XmlPart):
         duration = element.text
         return Duration.decode(duration)
 
-    def set_editing_duration(self, duration):
+    def set_editing_duration(self, duration: timedelta) -> None:
         """Set the time the document was edited.
 
+        (Also available as "self.editing_duration" property.)
+
         Arguments:
 
             duration -- timedelta
@@ -347,10 +539,25 @@ class Meta(XmlPart):
             self.get_meta_body().append(element)
         element.text = Duration.encode(duration)
 
-    def get_editing_cycles(self):
+    @property
+    def editing_duration(self) -> timedelta | None:
+        """Get or set the total time spent editing a document
+        <meta:editing-duration>.
+
+        Return: timedelta (or None if inexistant)
+        """
+        return self.get_editing_duration()
+
+    @editing_duration.setter
+    def editing_duration(self, duration: timedelta) -> None:
+        return self.set_editing_duration(duration)
+
+    def get_editing_cycles(self) -> int | None:
         """Get the number of times the document was edited, as reported by
         the generator.
 
+        (Also available as "self.editing_cycles" property.)
+
         Return: int (or None if inexistant)
         """
         element = self.get_element("//meta:editing-cycles")
@@ -359,9 +566,11 @@ class Meta(XmlPart):
         cycles = element.text
         return int(cycles)
 
-    def set_editing_cycles(self, cycles):
+    def set_editing_cycles(self, cycles: int) -> None:
         """Set the number of times the document was edited.
 
+        (Also available as "self.editing_cycles" property.)
+
         Arguments:
 
             cycles -- int
@@ -376,32 +585,42 @@ class Meta(XmlPart):
             self.get_meta_body().append(element)
         element.text = str(cycles)
 
-    def get_generator(self):
-        """Get the signature of the software that generated this document.
+    @property
+    def editing_cycles(self) -> int | None:
+        """Get or set the number of times a document has been edited
+        <meta:editing-cycles>.
+
+        When a document is created, this value is set to 1. Each time
+        a document is saved, the editing-cycles number is incremented by 1.
+
+        Return: int (or None if inexistant)
+        """
+        return self.get_editing_cycles()
+
+    @editing_cycles.setter
+    def editing_cycles(self, cycles: int) -> None:
+        return self.set_editing_cycles(cycles)
+
+    @property
+    def generator(self) -> str | None:
+        """Get or set the signature of the software that generated this
+        document.
 
         Return: str (or None if inexistant)
 
         Example::
 
-            >>> document.get_generator()
+            >>> document.meta.generator
             KOffice/2.0.0
+            >>> document.meta.generator = "Odfdo experiment"
         """
         element = self.get_element("//meta:generator")
         if element is None:
             return None
         return element.text
 
-    def set_generator(self, generator):
-        """Set the signature of the software that generated this document.
-
-        Arguments:
-
-            generator -- str
-
-        Example::
-
-            >>> document.set_generator(u"lpOD Project")
-        """
+    @generator.setter
+    def generator(self, generator: str) -> None:
         element = self.get_element("//meta:generator")
         if element is None:
             element = Element.from_tag("meta:generator")
@@ -409,19 +628,50 @@ class Meta(XmlPart):
         element.text = generator
         self._generator_modified = True
 
-    def set_generator_default(self):
+    def get_generator(self) -> str | None:
+        """Get the signature of the software that generated this document.
+
+        (Also available as "self.generator" property.)
+
+        Return: str (or None if inexistant)
+
+        Example::
+
+            >>> document.meta.get_generator()
+            KOffice/2.0.0
+        """
+        return self.generator
+
+    def set_generator(self, generator: str) -> None:
+        """Set the signature of the software that generated this document.
+
+        (Also available as "self.generator" property.)
+
+        Arguments:
+
+            generator -- str
+
+        Example::
+
+            >>> document.meta.set_generator("Odfdo experiment")
+        """
+        self.generator = generator
+
+    def set_generator_default(self) -> None:
         """Set the signature of the software that generated this document
         to ourself.
 
         Example::
 
-            >>> document.set_generator_default()
+            >>> document.meta.set_generator_default()
         """
         if not self._generator_modified:
-            self.set_generator("odfdo %s" % __version__)
+            self.generator = GENERATOR
 
-    def get_statistic(self):
-        """Get the statistic from the software that generated this document.
+    def get_statistic(self) -> dict[str, int] | None:
+        """Get the statistics about a document.
+
+        (Also available as "self.statistic" property.)
 
         Return: dict (or None if inexistant)
 
@@ -444,9 +694,10 @@ class Meta(XmlPart):
             statistic[to_str(key)] = int(value)
         return statistic
 
-    def set_statistic(self, statistic):
-        """Set the statistic for the documents: number of words, paragraphs,
-        etc.
+    def set_statistic(self, statistic: dict[str, int]) -> None:
+        """Set the statistics about a document.
+
+        (Also available as "self.statistic" property.)
 
         Arguments:
 
@@ -461,32 +712,89 @@ class Meta(XmlPart):
                              'meta:paragraph-count': 5,
                              'meta:word-count': 6,
                              'meta:character-count': 7}
-            >>> document.set_statistic(statistic)
+            >>> document.meta.set_statistic(statistic)
         """
         if not isinstance(statistic, dict):
-            raise TypeError("statistic must be a dict")
+            raise TypeError("Statistic must be a dict")
         element = self.get_element("//meta:document-statistic")
         for key, value in statistic.items():
             try:
                 ivalue = int(value)
-            except ValueError:
-                raise TypeError("statistic value must be a int")
+            except ValueError as e:
+                raise TypeError("Statistic value must be a int") from e
             element.set_attribute(to_str(key), str(ivalue))
 
-    def get_user_defined_metadata(self):
-        """Return a dict of str/value mapping.
+    @property
+    def statistic(self) -> dict[str, int] | None:
+        """Get or set the statistics about a document
+        <meta:document-statistic>.
+
+        Return: dict (or None if inexistant)
+
+        Example::
+
+            >>> document.get_statistic():
+            {'meta:table-count': 1,
+             'meta:image-count': 2,
+             'meta:object-count': 3,
+             'meta:page-count': 4,
+             'meta:paragraph-count': 5,
+             'meta:word-count': 6,
+             'meta:character-count': 7}
+        """
+        return self.get_statistic()
+
+    @statistic.setter
+    def statistic(self, statistic: dict[str, int]) -> None:
+        return self.set_statistic(statistic)
+
+    def get_user_defined_metadata(self) -> dict[str, Any]:
+        """Get all additional user-defined metadata for a document.
+
+        (Also available as "self.user_defined_metadata" property.)
+
+        Return a dict of str/value mapping.
 
         Value types can be: Decimal, date, time, boolean or str.
         """
-        result = {}
+        result: dict[str, Any] = {}
         for item in self.get_elements("//meta:user-defined"):
+            if not isinstance(item, Element):
+                continue
             # Read the values
-            name = item.get_attribute("meta:name")
+            name = item.get_attribute_string("meta:name")
+            if name is None:
+                continue
             value = self._get_meta_value(item)
             result[name] = value
         return result
 
-    def get_user_defined_metadata_of_name(self, keyname):
+    def clear_user_defined_metadata(self) -> None:
+        """Remove all user-defined metadata."""
+        while True:
+            element = self.get_element("//meta:user-defined")
+            if isinstance(element, Element):
+                element.delete()
+                continue
+            break
+
+    @property
+    def user_defined_metadata(self) -> dict[str, Any]:
+        """Get or set all additional user-defined metadata for a document.
+
+        Return a dict of str/value mapping.
+
+        Value types can be: Decimal, date, time, boolean or str.
+        """
+        return self.get_user_defined_metadata()
+
+    @user_defined_metadata.setter
+    def user_defined_metadata(self, metadata: dict[str, Any]) -> None:
+        self.clear_user_defined_metadata()
+        for key, val in metadata.items():
+            self.set_user_defined_metadata(name=key, value=val)
+
+    def get_user_defined_metadata_of_name(self, keyname: str) -> dict[str, Any] | None:
         """Return the content of the user defined metadata of that name.
         Return None if no name matchs or a dic of fields.
 
@@ -497,6 +805,8 @@ class Meta(XmlPart):
         result = {}
         found = False
         for item in self.get_elements("//meta:user-defined"):
+            if not isinstance(item, Element):
+                continue
             # Read the values
             name = item.get_attribute("meta:name")
             if name == keyname:
@@ -505,13 +815,13 @@ class Meta(XmlPart):
         if not found:
             return None
         result["name"] = name
-        value, value_type, text = self._get_meta_value(item, full=True)
+        value, value_type, text = self._get_meta_value(item, full=True)  # type: ignore
         result["value"] = value
         result["value_type"] = value_type
         result["text"] = text
         return result
 
-    def set_user_defined_metadata(self, name, value):
+    def set_user_defined_metadata(self, name: str, value: Any) -> None:
         if isinstance(value, bool):
             value_type = "boolean"
             value = "true" if value else "false"
@@ -530,9 +840,11 @@ class Meta(XmlPart):
             value_type = "time"
             value = str(Duration.encode(value))
         else:
-            raise TypeError('unexpected type "%s" for value' % type(value))
+            raise TypeError(f'unexpected type "{type(value)}" for value')
         # Already the same element ?
         for metadata in self.get_elements("//meta:user-defined"):
+            if not isinstance(metadata, Element):
+                continue
             if metadata.get_attribute("meta:name") == name:
                 break
         else:
@@ -542,29 +854,35 @@ class Meta(XmlPart):
         metadata.set_attribute("meta:value-type", value_type)
         metadata.text = value
 
+    def _get_meta_value(
+        self, element: Element, full: bool = False
+    ) -> Any | tuple[Any, str, str]:
+        """get_value() deicated to the meta data part, for one meta element."""
+        if full:
+            return self._get_meta_value_full(element)
+        else:
+            return self._get_meta_value_full(element)[0]
+
     @staticmethod
-    def _get_meta_value(element, full=False):
+    def _get_meta_value_full(element: Element) -> tuple[Any, str, str]:
         """get_value deicated to the meta data part, for one meta element."""
         # name = element.get_attribute('meta:name')
-        value_type = element.get_attribute("meta:value-type")
+        value_type = element.get_attribute_string("meta:value-type")
         if value_type is None:
             value_type = "string"
         text = element.text
         # Interpretation
         if value_type == "boolean":
-            value = Boolean.decode(text)
-        elif value_type in ("float", "percentage", "currency"):
-            value = Decimal(text)
-        elif value_type == "date":
+            return (Boolean.decode(text), value_type, text)
+        if value_type in ("float", "percentage", "currency"):
+            return (Decimal(text), value_type, text)
+        if value_type == "date":
             if "T" in text:
-                value = DateTime.decode(text)
+                return (DateTime.decode(text), value_type, text)
             else:
-                value = Date.decode(text)
-        elif value_type == "string":
-            value = text
-        elif value_type == "time":
-            value = Duration.decode(text)
-        if full:
-            return (value, value_type, text)
-        else:
-            return value
+                return (Date.decode(text), value_type, text)
+        if value_type == "string":
+            return (text, value_type, text)
+        if value_type == "time":
+            return (Duration.decode(text), value_type, text)
+        raise TypeError(f"Unknown value type: '{value_type!r}'")
