@@ -1,4 +1,4 @@
-# Copyright 2018-2022 Jérôme Dumonteil
+# Copyright 2018-2024 Jérôme Dumonteil
 # Copyright (c) 2009-2012 Ars Aperta, Itaapy, Pierlis, Talend.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -18,6 +18,15 @@
 # The odfdo project is a derivative work of the lpod-python project:
 # https://github.com/lpod/lpod-python
 # Authors: Jerome Dumonteil <jerome.dumonteil@itaapy.com>
-"""Version number of the package
+"""Version number of the package.
 """
-__version__ = "3.3.2"
+from __future__ import annotations
+
+try:
+    from importlib.metadata import version as __version
+except (AttributeError, ImportError):  # pragma: no cover
+    import importlib
+
+    __version = importlib.metadata.version
+
+__version__ = __version("odfdo")
